
#ifndef interpreter_h_INCLUDED 
#define interpreter_h_INCLUDED

#include "screen_interface.h"
#include "../fizmo/fizmo.h"

void interpreter_register_screen_interface(
    struct z_screen_interface *screen_interface);
int fizmo_parse_commandline_parameters(int argc, char *argv[], int argi);
void fizmo_start(char* z_story_filename);
void fizmo_new_screen_size(int width, int height);

#endif /* interpreter_h_INCLUDED */

