
#ifndef mt19937ar_h_INCLUDED 
#define mt19937ar_h_INCLUDED

#define GENRAND_INT32_MAX 0xffffffff

/* initializes mt[N] with a seed */
void init_genrand(unsigned long s);

/* initialize by an array with array-length */
/* init_key is the array for initializing keys */
/* key_length is its length */
/* slight change for C++, 2004/2/26 */
void init_by_array(unsigned long init_key[], int key_length);

/* generates a random number on [0,0xffffffff]-interval */
unsigned long genrand_int32(void);

#endif /* mt19937ar_h_INCLUDED */

