
#ifndef undo_h_INCLUDED
#define undo_h_INCLUDED

#include "types.h"

void opcode_save_undo(void);
void opcode_restore_undo(void);
size_t get_allocated_undo_memory_size(void);

#endif /* undo_h_INCLUDED */

