
#ifndef z_ucs_i_h_INCLUDED
#define z_ucs_i_h_INCLUDED

#include "types.h"

// z_ucs-Functions using i18n calls are grouped into this header file. This
// allowed prepi18n to use some z_ucs-function without having to include every
// last dependency.

char *z_ucs_to_system_charset(z_ucs *string);

#endif // z_ucs_i_h_INCLUDED

