
#ifndef output_h_INCLUDED
#define output_h_INCLUDED

#define STYLE_ROMAN 0
#define STYLE_REVERSE_VIDEO 1
#define STYLE_BOLD 2
#define STYLE_ITALIC 4
#define STYLE_FIXED_PITCH 8

#ifndef output_c_INCLUDED
/*@-exportlocal@*/
extern int16_t active_window_number;
extern z_font current_font;
extern bool lower_window_buffering_active;
extern z_style current_style;
extern z_style upper_window_style;
extern z_style lower_window_style;
extern z_colour current_foreground_colour;
extern z_colour current_background_colour;
extern z_colour upper_window_foreground_colour;
extern z_colour upper_window_background_colour;
extern z_colour lower_window_foreground_colour;
extern z_colour lower_window_background_colour;
/*@+exportlocal@*/
#endif /* output_c_INCLUDED */

void opcode_split_window(void);
void opcode_set_window(void);
void opcode_erase_window(void);
void opcode_erase_line_value(void);
void opcode_erase_line_pixels(void);
void opcode_set_cursor(void);
void opcode_get_cursor(void);
void opcode_set_text_style(void);
void opcode_buffer_mode(void);
void opcode_set_colour_fb(void);
void opcode_set_colour_fbw(void);
void opcode_set_font(void);

#endif /* output_h_INCLUDED */

