
#ifndef i18n_fizmo_codes_h_INCLUDED
#define i18n_fizmo_codes_h_INCLUDED

#define i18n_fizmo_RANDOM_GENERATOR_IS_NOW_IN_PREDICTABLE_MODE 1
#define i18n_fizmo_RANDOM_GENERATOR_IS_NOW_IN_RANDOM_MODE 2
#define i18n_fizmo_FUNCTION_CALL_MALLOC_P0D_RETURNED_NULL_PROBABLY_OUT_OF_MEMORY 3
#define i18n_fizmo_COULD_NOT_OPEN_FILE_NAMED_P0S 4
#define i18n_fizmo_COULD_NOT_OPEN_TRACE_FILE_P0S 5
#define i18n_fizmo_FUNCTION_CALL_REALLOC_P0D_RETURNED_NULL_PROBABLY_OUT_OF_MEMORY 6
#define i18n_fizmo_ERROR_READING_FIRST_STORY_BYTE_FROM_P0S 7
#define i18n_fizmo_UNKNOWN_STORY_VERSION_P0D 8
#define i18n_fizmo_INVALID_PARAMETER_TYPE_P0S 9
#define i18n_fizmo_INVALID_IFF_ACCESS_MODE_P0D 10
#define i18n_fizmo_OPCODE_JE_WITH_ONLY_1_OPERAND_IS_ILLEGAL 11
#define i18n_fizmo_FUNCTION_CALL_TIME_RETURNED_NEG_1 12
#define i18n_fizmo_FUNCTION_CALL_GETTIMEOFDAY_RETURNED_NONZERO 13
#define i18n_fizmo_CANNOT_DIVIDE_BY_ZERO 14
#define i18n_fizmo_COULD_NOT_FIND_ORIGINAL_STORY_FILE_P0S 15
#define i18n_fizmo_NOT_YET_IMPLEMENTED 16
#define i18n_fizmo_OBJECT_NUMBER_P0D_NOT_ALLOWED_IN_STORY_VERSION_P1D 17
#define i18n_fizmo_ATTRIBUTE_NUMBER_P0D_NOT_ALLOWED_IN_STORY_VERSION_P1D 18
#define i18n_fizmo_ERROR_WHILE_READING_FILE_P0S 19
#define i18n_fizmo_ERROR_WHILE_CLOSING_FILE_P0S 20
#define i18n_fizmo_STORY_FILE_OF_SIZE_P0D_TOO_LARGE_ONLY_P1D_BYTES_ALLOWED 21
#define i18n_fizmo_NO_PROPERTY_P0D_FOR_OBJECT_P1D 22
#define i18n_fizmo_CANNOT_READ_PROPERTIES_WITH_A_LENGTH_GREATER_THAN_2 23
#define i18n_fizmo_MORE_THAN_15_LOCALS_ARE_NOT_ALLOWED 24
#define i18n_fizmo_FATAL_ERROR_READING_STORY_FILE 25
#define i18n_fizmo_CANT_FIND_CHUNK_IFHD 26
#define i18n_fizmo_COULD_NOT_READ_RELEASE_NUMBER 27
#define i18n_fizmo_COULD_NOT_READ_SERIAL_NUMBER 28
#define i18n_fizmo_COULD_NOT_READ_CHECKSUM 29
#define i18n_fizmo_COULD_NOT_READ_RESTORE_PC 30
#define i18n_fizmo_CANT_FIND_CMEM_OR_UMEM_CHUNK 31
#define i18n_fizmo_CANT_READ_CHUNK_LENGTH 32
#define i18n_fizmo_ERROR_READING_SAVE_FILE 33
#define i18n_fizmo_CANT_FIND_CHUNK_STKS 34
#define i18n_fizmo_CANNOT_PULL_FROM_EMPTY_STACK 35
#define i18n_fizmo_CANNOT_DROP_P0D_WORDS_FROM_STACK_NOT_ENOUGH_WORDS_STORED 36
#define i18n_fizmo_UNKNOWN_CHAR_CODE_P0D 37
#define i18n_fizmo_INVALID_ZSCII_OUTPUT_CODE_P0D 38
#define i18n_fizmo_UNKNOWN_ERROR_CASE 39
#define i18n_fizmo_MAXIMUM_NUMBER_OF_STACK_ENTRIES_PER_ROUTINE_P0D_EXCEEDED 40
#define i18n_fizmo_TRYING_TO_STORE_VARIABLE_L_P0D_BUT_ONLY_P1D_VARIABLES_ACTIVE 41
#define i18n_fizmo_NOT_ENOUGH_STACK_WORDS_FROM_LOCAL_ROUTINE_ON_STACK 42
#define i18n_fizmo_INSTRUCTION_FORM_NOT_INITIALIZED 43
#define i18n_fizmo_UNKNOWN_OPERAND_TYPE_P0D 44
#define i18n_fizmo_OPCODE_P0D_IN_FORM_P1D_NOT_IMPLEMENTED 45
#define i18n_fizmo_OBJECT_NUMBER_0_IS_NOT_VALID 46
#define i18n_fizmo_MAXIMUM_ABBREVIATION_DEPTH_IS_P0D 47
#define i18n_fizmo_PLEASE_ENTER_SCRIPT_FILENAME 48
#define i18n_fizmo_INVALID_BACKSLASH_SEQUENCE_IN_LOCALIZATION_DATA 49
#define i18n_fizmo_INVALID_OUTPUT_STREAM_NUMBER_P0D 50
#define i18n_fizmo_WARNING_FOR_P0S_AT_P0X 51
#define i18n_fizmo_NULL_POINTER_RECEIVED 52
#define i18n_fizmo_INVALID_NODE_TYPE_P0D 53
#define i18n_fizmo_PROPERTY_NUMBER_0_IS_NOT_VALID 54
#define i18n_fizmo_PROPERTY_NUMBER_P0D_NOT_ALLOWED_IN_STORY_VERSION_P1D 55
#define i18n_fizmo_CAUGHT_SIGNAL_P0D_ABORTING_INTERPRETER 56
#define i18n_fizmo_TIMED_INPUT_NOT_IMPLEMENTED_IN_INTERFACE_P0S 57
#define i18n_fizmo_FUNCTION_CALL_P0S_RETURNED_ERROR_P1D_P2S 58
#define i18n_fizmo_ERROR_P0D_OCCURED_BEFORE_READ_P0S 59
#define i18n_fizmo_PRELOADED_INPUT_NOT_AVAILIABLE_IN_INTERFACE_P0S 60
#define i18n_fizmo_FUNCTION_CALL_P0S_RETURNED_ERROR_CODE_P1D 61
#define i18n_fizmo_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR 62
#define i18n_fizmo_PLEASE_ENTER_SAVEGAME_FILENAME 63
#define i18n_fizmo_FIZMO_VERSION_P0S 64
#define i18n_fizmo_STORY_RELEASE_NUMBER 65
#define i18n_fizmo_STORY_SERIAL_NUMBER 66
#define i18n_fizmo_CURRENT_Z_STACK_SIZE_P0D_ENTRIES 67
#define i18n_fizmo_HACK_01__ROUTINE_STACK_UNDERFLOW_CHECK_DISABLED 68
#define i18n_fizmo_VERSION_P0D_GAME_NOT_SUPPORTED_BY_P1S_INTERFACE 69
#define i18n_fizmo_CURRENT_Z_STACK_ENTRIES_IN_USE_P0D 70
#define i18n_fizmo_INVALID_THROW_DESTINATION_STACK_INDEX_P0D 71
#define i18n_fizmo_MAXIMUM_STREAM_3_DEPTH_P0D_EXCEEDED 72
#define i18n_fizmo_INVALID_ZSCII_INPUT_CODE_P0D 73
#define i18n_fizmo_ERROR_WRITING_SAVE_FILE 74
#define i18n_fizmo_P0D_BYTES_USED_BY_TEXT_HISTORY 75
#define i18n_fizmo_P0D_BYTES_USED_BY_BLOCK_BUFFER 76
#define i18n_fizmo_P0D_BYTES_USED_BY_TEXT_HISTORY_METADATA 77
#define i18n_fizmo_VALID_COMMANDS_ARE 78
#define i18n_fizmo_P0D_BYTES_USED_BY_PARAGRAPH_CACHE 79
#define i18n_fizmo_RELEASE_NR_SERIAL_NR_OR_CHECKSUM_DOESNT_MATCH 80
#define i18n_fizmo_PLEASE_ENTER_NAME_FOR_COMMANDFILE 81
#define i18n_fizmo_FILENAME_MUST_NOT_BE_EMPTY 82
#define i18n_fizmo_THIS_FUNCTION_HAS_BEEN_DISABLED 83
#define i18n_fizmo_P0D_BYTES_USED_FOR_UNDO 84
#define i18n_fizmo_UNKNOWN_CONFIGURATION_OPTION_P0S 85

#endif /* i18n_fizmo_codes_h_INCLUDED */

