
#ifndef i18n_ncursesw_codes_h_INCLUDED
#define i18n_ncursesw_codes_h_INCLUDED

#define i18n_ncursesw_SET_FOREGROUND_COLOR 1
#define i18n_ncursesw_SET_BACKGROUND_COLOR 2
#define i18n_ncursesw_SET_LEFT_MARGIN_SIZE 3
#define i18n_ncursesw_SET_RIGHT_MARGIN_SIZE 4
#define i18n_ncursesw_INVALID_CONFIGURATION_VALUE_P0S_FOR_P1S 5
#define i18n_ncursesw_USAGE_DESCRIPTION 6
#define i18n_ncursesw_VALID_OPTIONS_ARE 7
#define i18n_ncursesw_SET_LOCALE_NAME_FOR_INTERPRETER_MESSAGES 8
#define i18n_ncursesw_START_WITH_RANDOM_GENERATOR_IN_PREDICTABLE_MODE 9
#define i18n_ncursesw_START_GAME_WITH_TRANSCRIPT_ENABLED 10
#define i18n_ncursesw_FIZMO_VERSION_P0S 11
#define i18n_ncursesw_LOCALES_AVAILIABLE 12
#define i18n_ncursesw_MORE_PROMPT 13
#define i18n_ncursesw_PRESS_ANY_KEY_TO_QUIT 14
#define i18n_ncursesw_SCORE 15
#define i18n_ncursesw_TURNS 16
#define i18n_ncursesw_NUMBER_OF_COLOR_COMBINATIONS_AVAILABLE_P0D 17
#define i18n_ncursesw_COLORS_AVAILABLE 18
#define i18n_ncursesw_INVALID_COLOR_NAME 19
#define i18n_ncursesw_SHOW_HELP_MESSAGE_AND_EXIT 20
#define i18n_ncursesw_USE_BOLD_FOR_BRIGHT_FOREGROUND_COLORS 21
#define i18n_ncursesw_USE_BLINK_FOR_BRIGHT_BACKGROUND_COLORS 22
#define i18n_ncursesw_DONT_USE_COLORS 23
#define i18n_ncursesw_ENABLE_COLORS 24
#define i18n_ncursesw_NO_STORIES_REGISTERED_PLUS_HELP 25
#define i18n_ncursesw_SPACE_FOR_NEXT_PAGE 26
#define i18n_ncursesw_B_FOR_LAST_PAGE 27
#define i18n_ncursesw_USE_UMEM_FOR_SAVEGAMES 28
#define i18n_ncursesw_ENABLE_XTERM_GRAPHICS 29
#define i18n_ncursesw_USE_XTERM_TITLE 30
#define i18n_ncursesw_FORCE_8BIT_SOUND 31
#define i18n_ncursesw_DISABLE_SOUND 32
#define i18n_ncursesw_SET_TANDY_FLAG 33
#define i18n_ncursesw_DONT_UPDATE_STORY_LIST_ON_START 34
#define i18n_ncursesw_SEARCH_DIRECTORY 35
#define i18n_ncursesw_RECURSIVELY_SEARCH_DIRECTORY 36
#define i18n_ncursesw_START_GAME_WITH_RECORDING_COMMANDS 37
#define i18n_ncursesw_START_GAME_WITH_INPUT_FROM_FILE 38
#define i18n_ncursesw_UPDATE_STORY_LIST_ON_START 39
#define i18n_ncursesw_FORCE_RANDOM_GENERATOR_IN_PREDICTABLE_MODE 40

#endif /* i18n_ncursesw_codes_h_INCLUDED */

