
#ifndef i18n_codes_h_INCLUDED
#define i18n_codes_h_INCLUDED

#define i18n_USAGE_DESCRIPTION 1
#define i18n_RANDOM_GENERATOR_IS_NOW_IN_PREDICTABLE_MODE 2
#define i18n_RANDOM_GENERATOR_IS_NOW_IN_RANDOM_MODE 3
#define i18n_FUNCTION_CALL_MALLOC_P0D_RETURNED_NULL_PROBABLY_OUT_OF_MEMORY 4
#define i18n_COULD_NOT_OPEN_FILE_NAMED_P0S 5
#define i18n_COULD_NOT_OPEN_TRACE_FILE_P0S 6
#define i18n_FUNCTION_CALL_REALLOC_P0D_RETURNED_NULL_PROBABLY_OUT_OF_MEMORY 7
#define i18n_ERROR_READING_FIRST_STORY_BYTE_FROM_P0S 8
#define i18n_UNKNOWN_STORY_VERSION_P0D 9
#define i18n_INVALID_OUTPUT_ENCODING_P0D 10
#define i18n_INVALID_PARAMETER_TYPE_P0S 11
#define i18n_INVALID_IFF_ACCESS_MODE_P0D 12
#define i18n_OPCODE_JE_WITH_ONLY_1_OPERAND_IS_ILLEGAL 13
#define i18n_FUNCTION_CALL_TIME_RETURNED_NEG_1 14
#define i18n_FUNCTION_CALL_GETTIMEOFDAY_RETURNED_NONZERO 15
#define i18n_CANNOT_DIVIDE_BY_ZERO 16
#define i18n_INVALID_RANDOM_MODE_P0D 17
#define i18n_COULD_NOT_FIND_ORIGINAL_STORY_FILE_P0S 18
#define i18n_NOT_YET_IMPLEMENTED 19
#define i18n_OBJECT_NUMBER_P0D_NOT_ALLOWED_IN_STORY_VERSION_P1D 20
#define i18n_ATTRIBUTE_NUMBER_P0D_NOT_ALLOWED_IN_STORY_VERSION_P1D 21
#define i18n_ERROR_WHILE_READING_FILE_P0S 22
#define i18n_ERROR_WHILE_CLOSING_FILE_P0S 23
#define i18n_STORY_FILE_OF_SIZE_P0D_TOO_LARGE_ONLY_P1D_BYTES_ALLOWED 24
#define i18n_NO_PROPERTY_P0D_FOR_OBJECT_P1D 25
#define i18n_CANNOT_READ_PROPERTIES_WITH_A_LENGTH_GREATER_THAN_2 26
#define i18n_MORE_THAN_15_LOCALS_ARE_NOT_ALLOWED 27
#define i18n_FATAL_ERROR_READING_STORY_FILE 28
#define i18n_CANT_FIND_CHUNK_IFHD 29
#define i18n_COULD_NOT_READ_RELEASE_NUMBER 30
#define i18n_COULD_NOT_READ_SERIAL_NUMBER 31
#define i18n_COULD_NOT_READ_CHECKSUM 32
#define i18n_COULD_NOT_READ_RESTORE_PC 33
#define i18n_CANT_FIND_CHUNK_CMEM 34
#define i18n_CANT_READ_CHUNK_LENGTH 35
#define i18n_ERROR_READING_SAVE_FILE 36
#define i18n_CANT_FIND_CHUNK_STKS 37
#define i18n_CANNOT_PULL_FROM_EMPTY_STACK 38
#define i18n_CANNOT_DROP_P0D_WORDS_FROM_STACK_NOT_ENOUGH_WORDS_STORED 39
#define i18n_UNKNOWN_CHAR_CODE_P0D 40
#define i18n_INVALID_ZSCII_OUTPUT_CODE_P0D 41
#define i18n_UNKNOWN_ERROR_CASE 42
#define i18n_INVALID_UTF_CHAR_CODE_P0D 43
#define i18n_MAXIMUM_NUMBER_OF_STACK_ENTRIES_PER_ROUTINE_P0D_EXCEEDED 44
#define i18n_TRYING_TO_STORE_VARIABLE_L_P0D_BUT_ONLY_P1D_VARIABLES_ACTIVE 45
#define i18n_NOT_ENOUGH_STACK_WORDS_FROM_LOCAL_ROUTINE_ON_STACK 46
#define i18n_INSTRUCTION_FORM_NOT_INITIALIZED 47
#define i18n_UNKNOWN_OPERAND_TYPE_P0D 48
#define i18n_OPCODE_P0D_NOT_IMPLEMENTED 49
#define i18n_OBJECT_NUMBER_0_IS_NOT_VALID 50
#define i18n_MAXIMUM_ABBREVIATION_DEPTH_IS_P0D 51
#define i18n_PLEASE_ENTER_SCRIPT_FILENAME_OR_LEAVE_EMPTY_TO_ACCEPT_P0S 52
#define i18n_INVALID_BACKSLASH_SEQUENCE_IN_LOCALIZATION_DATA 53
#define i18n_INVALID_OUTPUT_STREAM_NUMBER_P0D 54
#define i18n_WARNING_FOR_P0S_AT_P0X 55
#define i18n_NULL_POINTER_RECEIVED 56
#define i18n_INVALID_NODE_TYPE_P0D 57
#define i18n_PROPERTY_NUMBER_0_IS_NOT_VALID 58
#define i18n_PROPERTY_NUMBER_P0D_NOT_ALLOWED_IN_STORY_VERSION_P1D 59
#define i18n_CAUGHT_SIGNAL_P0D_ABORTING_INTERPRETER 60
#define i18n_COULD_NOT_FIND_REQUESTED_INTERFACE_P0S 61
#define i18n_TIMED_INPUT_NOT_IMPLEMENTED_IN_INTERFACE_P0S 62
#define i18n_FUNCTION_CALL_P0S_RETURNED_ERROR_P1D_P2S 63
#define i18n_MORE_PROMPT 64
#define i18n_ERROR_P0D_OCCURED_BEFORE_READ_P0S 65
#define i18n_PRELOADED_INPUT_NOT_AVAILIABLE_IN_INTERFACE_P0S 66
#define i18n_FUNCTION_CALL_P0S_RETURNED_ERROR_CODE_P1D 67
#define i18n_FUNCTION_CALL_P0S_ABORTED_DUE_TO_ERROR 68
#define i18n_PRESS_ANY_KEY_TO_QUIT 69
#define i18n_PLEASE_ENTER_SAVEGAME_FILENAME_OR_LEAVE_EMPTY_TO_ACCEPT_P0Z 70
#define i18n_C_IF_ECHO_PARAMETER_DESCRIPTION 71
#define i18n_C_IF_XTERM_STATUS_LINE_DESCRIPTION 72
#define i18n_VALID_OPTIONS_ARE 73
#define i18n_USE_PLAIN_ASCII_OUTPUT 74
#define i18n_SPECIFY_ISO_8859_1_CAPABLE_OUTPUT_DEVICE 75
#define i18n_SET_LOCALE_NAME_FOR_INTERPRETER_MESSAGES 76
#define i18n_START_WITH_RANDOM_GENERATOR_IN_PREDICTABLE_MODE 77
#define i18n_START_GAME_WITH_TRANSCRIPT_ENABLED 78
#define i18n_SPECIFY_UTF_8_CAPABLE_OUTPUT_DEVICE 79
#define i18n_LOCALES_AVAILIABLE 80
#define i18n_DISABLE_INFOCOM_COMPATIBILITY_HACKS 81
#define i18n_FIZMO_VERSION_P0S 82
#define i18n_STORY_RELEASE_NUMBER 83
#define i18n_STORY_SERIAL_NUMBER 84
#define i18n_CURRENT_Z_STACK_SIZE_P0D_ENTRIES 85
#define i18n_HACK_01__ROUTINE_STACK_UNDERFLOW_CHECK_DISABLED 86
#define i18n_MAXIMUM_NUMBER_OF_Z_WINDOWS_EXCEEDED 87
#define i18n_VERSION_P0D_GAME_NOT_SUPPORTED_BY_P1S_INTERFACE 88
#define i18n_CURRENT_Z_STACK_ENTRIES_IN_USE_P0D 89
#define i18n_INVALID_THROW_DESTINATION_STACK_INDEX_P0D 90
#define i18n_SCORE 91
#define i18n_TURNS 92
#define i18n_MAXIMUM_STREAM_3_DEPTH_P0D_EXCEEDED 93
#define i18n_INVALID_ZSCII_INPUT_CODE_P0D 94
#define i18n_ERROR_WRITING_SAVE_FILE 95

#define TOTAL_NUMBER_OF_I18N_MESSAGES 95

#endif /* i18n_codes_h_INCLUDED */

