
#ifndef property_h_INCLUDED
#define property_h_INCLUDED

#include <inttypes.h>

void opcode_get_prop(void);
void opcode_put_prop(void);
void opcode_get_prop_addr(void);
void opcode_get_prop_len(void);
void opcode_get_next_prop(void);

#endif /* property_h_INCLUDED */

