
#include <inttypes.h>

#ifndef math_h_INCLUDED
#define math_h_INCLUDED

void opcode_and(void);
void opcode_add(void);
void opcode_sub(void);
void opcode_je(void);
void opcode_jz(void);
void opcode_jg(void);
void opcode_inc_chk(void);
void opcode_test(void);
void opcode_jl(void);
void opcode_dec_chk(void);
void opcode_mul(void);
void opcode_random(void);
void opcode_div(void);
void opcode_or(void);
void opcode_mod(void);
void opcode_not(void);
void opcode_art_shift(void);
void opcode_log_shift(void);
void seed_random_generator(void);

#endif /* math_h_INCLUDED */

