
#ifndef zscii_h_INCLUDED
#define zscii_h_INCLUDED

#define ZSCII_NULL 0
#define ZSCII_DELETE 8
#define ZSCII_TAB 9
#define ZSCII_SENTENCE_SPACE 11
#define ZSCII_NEWLINE 13
#define ZSCII_ESCAPE 27
#define ZSCII_CURSOR_UP 129
#define ZSCII_CURSOR_DOWN 130
#define ZSCII_CURSOR_LEFT 131
#define ZSCII_CURSOR_RIGHT 132
#define ZSCII_F1 133
#define ZSCII_F2 134
#define ZSCII_F3 135
#define ZSCII_F4 136
#define ZSCII_F5 137
#define ZSCII_F6 138
#define ZSCII_F7 139
#define ZSCII_F8 140
#define ZSCII_F9 141
#define ZSCII_F10 142
#define ZSCII_F11 143
#define ZSCII_F12 144
#define ZSCII_KEYPAD_0 145
#define ZSCII_KEYPAD_1 146
#define ZSCII_KEYPAD_2 147
#define ZSCII_KEYPAD_3 148
#define ZSCII_KEYPAD_4 149
#define ZSCII_KEYPAD_5 150
#define ZSCII_KEYPAD_6 151
#define ZSCII_KEYPAD_7 152
#define ZSCII_KEYPAD_8 153
#define ZSCII_KEYPAD_9 154

#endif /* zscii_h_INCLUDED */

