
#include <iostream>
#include <stdlib.h>

#include "SimpleInterface.h"

#include "../cpp/ScreenInterface.h"
#include "../cpp/Interpreter.h"


char *SimpleInterface::get_interface_name()
{ return "simple-c++-interface"; }


int main(int argc, char *argv[])
{
  SimpleInterface *simple_interface = new SimpleInterface();
  Interpreter::register_screen_interface(simple_interface);

  if (argc != 2)
    exit(1);

  Interpreter::start(argv[1], NULL, NULL);

  return 0;
}

