
#ifndef splint_h_INCLUDED
#define splint_h_INCLUDED

#ifdef S_SPLINT_S

#define UINT32_MAX (4294967295L)
#define SIGQUIT 3
#define SIGBUS 10

extern char *strdup(char *s) /*@*/ ;

extern void TRACE_LOG(...);
extern void TRACE_LOG_Z_UCS(...);

#endif /* S_SPLINT_S */

#endif // splint_h_INCLUDED

