
#ifndef SimpleInterface_h_INCLUDED 
#define SimpleInterface_h_INCLUDED

#include "../cpp/ScreenInterface.h"


class SimpleInterface : public ScreenInterface
{
  //char *SimpleInterface::get_interface_name();
  char *get_interface_name();
};

#endif /* SimpleInterface_h_INCLUDED */

