
#ifndef variable_h_INCLUDED 
#define variable_h_INCLUDED

#include <inttypes.h>

#ifndef variable_c_INCLUDED 
/*@dependent@*/ extern uint16_t *local_variable_storage_index;
extern uint8_t number_of_locals_active;
#endif /* variable_c_INCLUDED */

void set_variable(uint8_t variable_number, uint16_t data);
uint16_t get_variable(uint8_t variable_number);
void opcode_pull(void);
void opcode_push(void);
void opcode_loadb(void);
void opcode_loadw(void);
void opcode_storew(void);
void opcode_store(void);
void opcode_inc(void);
void opcode_storeb(void);
void opcode_dec(void);
void opcode_load(void);
void opcode_pop(void);
void opcode_check_arg_count(void);

#endif /* variable_h_INCLUDED */

