
#ifndef object_h_INCLUDED
#define object_h_INCLUDED

#include <inttypes.h>

#define OBJECT_NODE_PARENT 0
#define OBJECT_NODE_SIBLING 1
#define OBJECT_NODE_CHILD 2

/*@dependent@*/ uint8_t *get_objects_property_table(uint16_t object_number);

void opcode_get_sibling(void);
void opcode_get_parent(void);
void opcode_get_child(void);
void opcode_jin(void);
void opcode_set_attr(void);
void opcode_test_attr(void);
void opcode_insert_obj(void);
void opcode_clear_attr(void);
void opcode_remove_obj(void);

#endif /* object_h_INCLUDED */

